# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------

from .base import Catalog
from .default import load_combo_catalog
from .local import EntrypointsCatalog, MergedCatalog


def _make_builtin():
    return MergedCatalog([EntrypointsCatalog(), load_combo_catalog()], name="builtin", description="Generated from data packages found on your intake search path")


def __getattr__(name):
    """Only make the builtin catalog on request"""
    global builtin
    if name == "builtin":
        builtin = _make_builtin()
        return builtin
    raise AttributeError(name)



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2019, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------
import panel as pn

from .base import ICONS, Base
from .catalog.gui import CatGUI
from .source.gui import SourceGUI


class GUI(Base):
    """
    Top level GUI panel that contains controls and all visible sub-panels

    This class is responsible for coordinating the inputs and outputs
    of various sup-panels and their effects on each other.

    Parameters
    ----------
    cats: list of catalogs
        catalogs used to initalize the cat panel

    Attributes
    ----------
    children: list of panel objects
        children that will be used to populate the panel when visible
    panel: panel layout object
        instance of a panel layout (row or column) that contains children
        when visible
    watchers: list of param watchers
        watchers that are set on children - cleaned up when visible
        is set to false.
    """

    def __init__(self, cats=None):
        self.source = SourceGUI()
        self.cat = CatGUI(cats=cats, done_callback=self.done_callback)
        self.panel = pn.Column(
            pn.Row(
                pn.panel(ICONS["logo"]),
                pn.Column(self.cat.select.panel, self.cat.control_panel, margin=0, width_policy="max"),
                pn.Column(self.source.select.panel, self.source.control_panel, margin=0, width_policy="max"),
                self.source.description.panel,
                margin=0,
                width_policy="max",
            ),
            pn.Row(self.cat.search.panel, self.cat.add.panel, self.source.plot.panel, width_policy="max"),
            width_policy="max",
        )
        super(GUI, self).__init__()

    def done_callback(self, cats):
        self.source.select.cats = cats

    @property
    def cats(self):
        """Cats that have been selected from the cat sub-panel"""
        return self.cat.cats

    def add(self, *args, **kwargs):
        """Add to list of cats"""
        return self.cat.select.add(*args, **kwargs)

    @property
    def sources(self):
        """Sources that have been selected from the source sub-panel"""
        return self.source.sources

    @property
    def source_instance(self):
        """DataSource instance for the current selection and any parameters"""
        return self.source.source_instance

    @property
    def item(self):
        """Item that is selected"""
        if len(self.sources) == 0:
            return None
        return self.sources[0]

    def __getstate__(self):
        """Serialize the current state of the object"""
        return {
            "visible": self.visible,
            "cat": self.cat.__getstate__(),
            "source": self.source.__getstate__(),
        }

    def __setstate__(self, state):
        """Set the current state of the object from the serialized version.
        Works inplace. See ``__getstate__`` to get serialized version and
        ``from_state`` to create a new object."""
        self.visible = state.get("visible", True)
        self.cat.__setstate__(state["cat"])
        self.source.__setstate__(state["source"])
        return self

    @classmethod
    def from_state(cls, state):
        """Create a new object from a serialized exising object.

        Example
        -------
        original = GUI()
        copy = GUI.from_state(original.__getstate__())
        """
        return cls(cats=[]).__setstate__(state)



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------

from intake.container.base import RemoteSource, get_partition
from intake.source.base import Schema


class RemoteSequenceSource(RemoteSource):
    """Sequence-of-things source on an Intake server"""

    name = "remote_sequence"
    container = "python"

    def __init__(self, url, headers, **kwargs):
        self.url = url
        self.npartitions = kwargs.get("npartition", 1)
        self.partition_access = self.npartitions > 1
        self.headers = headers
        self.metadata = kwargs.get("metadata", {})
        self._schema = Schema(npartitions=self.npartitions, extra_metadata=self.metadata)
        self.bag = None
        super(RemoteSequenceSource, self).__init__(url, headers, **kwargs)

    def _load_metadata(self):
        import dask
        import dask.bag as db

        if self.bag is None:
            self.parts = [dask.delayed(get_partition)(self.url, self.headers, self._source_id, self.container, i) for i in range(self.npartitions)]
            self.bag = db.from_delayed(self.parts)
        return self._schema

    def _get_partition(self, i):
        self._load_metadata()
        return self.parts[i].compute()

    def read(self):
        self._load_metadata()
        return self.bag.compute()

    def to_dask(self):
        self._load_metadata()
        return self.bag

    def _close(self):
        self.bag = None

    @staticmethod
    def _persist(source, path, encoder=None, **kwargs):
        """Save list to files using encoding

        encoder : None or one of str|json|pickle
            None is equivalent to str
        """
        import json
        import pickle

        encoder = {None: str, "str": str, "json": json.dumps, "pickle": pickle.dumps}[encoder]
        try:
            b = source.to_dask()
        except NotImplementedError:
            b = source.read()
        return RemoteSequenceSource._data_to_source(b, path, encoder, **kwargs)

    @staticmethod
    def _data_to_source(b, path, encoder=None, storage_options=None, **kwargs):
        import json
        import pickle
        import posixpath

        import dask
        import dask.bag as db
        from fsspec import open_files

        from intake.source.textfiles import TextFilesSource

        encoder = {None: str, "str": str, "json": json.dumps, "pickle": pickle.dumps}.get(encoder, encoder)

        if not hasattr(b, "to_textfiles"):
            try:
                b = db.from_sequence(b, npartitions=1)
            except TypeError:
                raise NotImplementedError

        files = open_files(posixpath.join(path, "part.*"), mode="wt", num=b.npartitions, **(storage_options or {}))
        dwrite = dask.delayed(write_file)
        out = [dwrite(part, f, encoder) for part, f in zip(b.to_delayed(), files)]
        dask.compute(out)
        s = TextFilesSource(posixpath.join(path, "part.*"), storage_options=storage_options)
        return s


def write_file(data, fo, encoder):
    with fo as f:
        for d in data:
            f.write(encoder(d))



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------

import itertools

from .base import RemoteSource, Schema, get_partition


class RemoteArray(RemoteSource):
    """nd-array on an Intake server"""

    name = "remote_ndarray"
    container = "ndarray"

    def __init__(self, url, headers, **kwargs):
        super(RemoteArray, self).__init__(url, headers, **kwargs)
        self.npartitions = kwargs["npartitions"]
        self.shape = tuple(kwargs["shape"])
        self.metadata = kwargs["metadata"]
        self.dtype = kwargs["dtype"]
        self.chunks = tuple(tuple(c) for c in tuple(kwargs.get("chunks", (-1,))))
        self.arr = None
        self._schema = Schema(npartitions=self.npartitions, extra_metadata=self.metadata, dtype=self.dtype, shape=self.shape, chunks=self.chunks)

    def _load_metadata(self):
        import dask.array as da

        if self.arr is None:
            name = "remote-array-" + self._source_id
            nparts = (range(len(n)) for n in self.chunks)
            dask = {(name,) + part: (get_partition, self.url, self.headers, self._source_id, self.container, part) for part in itertools.product(*nparts)}
            self.arr = da.Array(dask, name=name, chunks=self.chunks, dtype=self.dtype, shape=self.shape)

        return self._schema

    def _get_partition(self, i):
        self._load_metadata()
        return self.arr.blocks[i].compute()

    def read_partition(self, i):
        self._load_metadata()
        return self._get_partition(i)

    def read(self):
        self._load_metadata()
        return self.arr.compute()

    def to_dask(self):
        self._load_metadata()
        return self.arr

    def _close(self):
        self.arr = None

    @staticmethod
    def _persist(source, path, component=None, storage_options=None, **kwargs):
        """Save array to local persistent store

        Makes a parquet dataset out of the data using zarr.
        This then becomes a data entry in the persisted datasets catalog.
        Only works locally for the moment.

        Parameters
        ----------
        source: a DataSource instance to save
        name: str or None
            Key to refer to this persisted dataset by. If not given, will
            attempt to get from the source's name
        kwargs: passed on to zarr array creation, see
        """
        try:
            arr = source.to_dask()
        except NotImplementedError:
            arr = source.read()
        return RemoteArray._data_to_source(arr, path, component=None, storage_options=None, **kwargs)

    @staticmethod
    def _data_to_source(arr, path, component=None, storage_options=None, **kwargs):
        from dask.array import from_array, to_zarr
        from dask.utils import is_arraylike

        from ..source.zarr import ZarrArraySource

        if not is_arraylike(arr):
            raise NotImplementedError
        if not hasattr(arr, "npartitions"):
            arr = from_array(arr, chunks="auto")
        to_zarr(arr, path, component=None, storage_options=storage_options, **kwargs)
        source = ZarrArraySource(path, storage_options, component)
        return source



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------
""" Provide a ``main`` function to run intake commands.

"""

import logging

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------

# Standard library imports
import sys

# External imports

# Intake imports


# -----------------------------------------------------------------------------
# API
# -----------------------------------------------------------------------------


def die(message, status=1):
    """Print an error message and exit.
    This function will call ``sys.exit`` with the given ``status`` and the
    process will terminate.

    Args:
        message (str) : error message to print

        status (int) : the exit status to pass to ``sys.exit``

    """
    print(message, file=sys.stderr)
    sys.exit(status)


def nice_join(seq, sep=", ", conjunction="or"):
    """Join together sequences of strings into English-friendly phrases using
    a conjunction when appropriate.

    Args:
        seq (seq[str]) : a sequence of strings to nicely join

        sep (str, optional) : a sequence delimiter to use (default: ", ")

        conjunction (str or None, optional) : a conjunction to use for the last
            two items, or None to reproduce basic join behavior (default: "or")

    Returns:
        a joined string

    Examples:
        >>> nice_join(["a", "b", "c"])
        'a, b or c'

    """
    seq = [str(x) for x in seq]

    if len(seq) <= 1 or conjunction is None:
        return sep.join(seq)
    else:
        return "%s %s %s" % (sep.join(seq[:-1]), conjunction, seq[-1])


def print_entry_info(catalog, name):
    """ """
    info = catalog[name].describe()
    for key in sorted(info.keys()):
        print("[{}] {}={}".format(name, key, info[key]))


class Subcommand(object):
    """Abstract base class for subcommands

    Subclasses should define a class variable ``name`` that will be used as the
    subparser name, and a docstring, that will be used as the subparser help.
    After initialization, the parser for this comman will be avaialble as
    ``self.parser``.

    Subclasses must also implement:

    * an ``initialize(self)`` method that configures ``self.parser``

    * an ``invoke(self, args)`` method that accepts a set of argparse
      processed arguments as input.

    """

    def __init__(self, parser):
        """Configure a parser for this command."""
        self.parser = parser
        self.initialize()

    def initialize(self):
        """Implement in subclasses to configure self.parser with any arguments
        or additional sub-parsers.

        """
        raise NotImplementedError("Subclasses must implement initialize()")

    def invoke(self, args):
        """Implement in subclasses to perform the actual work of the command"""
        raise NotImplementedError("Subclasses must implement invoke()")



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------

from .cache import Cache
from .config import Config
from .describe import Describe
from .discover import Discover
from .drivers import Drivers
from .example import Example
from .exists import Exists
from .get import Get
from .info import Info
from .list import List
from .precache import Precache

all = (Cache, Config, Describe, Discover, Example, Exists, Get, Info, List, Precache, Drivers)



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------
"""

"""

import logging

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------

import sys

# Standard library imports
from importlib import import_module

# Intake imports
from intake import __version__
from intake.cli.util import Subcommand

# External imports


# -----------------------------------------------------------------------------
# API
# -----------------------------------------------------------------------------


class Info(Subcommand):
    """Display runtime information related to Intake"""

    name = "info"

    def initialize(self):
        pass

    def invoke(self, args):
        print("Python version      :  %s" % sys.version.split("\n")[0])
        print("IPython version     :  %s" % _version_from_module("IPython"))
        print("Tornado version     :  %s" % _version_from_module("tornado", "version"))
        print("Dask version        :  %s" % _version_from_module("dask"))
        print("Pandas version      :  %s" % _version_from_module("pandas"))
        print("Numpy version       :  %s" % _version_from_module("numpy"))
        print("Intake version      :  %s" % __version__)


def _version_from_module(modname, version_attr="__version__"):
    try:
        mod = import_module(modname)
        return getattr(mod, version_attr)
    except ImportError:
        return "(not installed)"



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2019, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------

try:
    import panel as pn

    from ..interface.gui import ICONS, SourceGUI

    css = """
    .scrolling {
        overflow: scroll;
    }
    """
    pn.config.raw_css.append(css)  # add scrolling class from css (panel GH#383, GH#384)
    pn.extension()

    class EntryGUI(SourceGUI):
        def __init__(self, source=None, **kwargs):
            self.source = source
            super().__init__(sources=[self.source], **kwargs)
            self.panel = pn.Column(pn.Row(ICONS["logo"], pn.Column(*self.controls)), self.description.panel, self.plot.panel)

        def setup(self):
            self._setup_watchers()

        @SourceGUI.visible.setter
        def visible(self, visible):
            self._visible = visible

            if visible:
                self.setup()
                self.select.visible = True
                self.description.visible = True
            elif not visible:
                self.unwatch()
                # do children
                self.select.visible = False
                self.description.visible = False
                self.plot.visible = False
            if self.visible_callback:
                self.visible_callback(visible)

        @property
        def item(self):
            return self.source

    class CatalogGUI(SourceGUI):
        def __init__(self, cat, **kwargs):
            self.cat = cat
            super().__init__(cats=[self.cat], **kwargs)
            self.panel = pn.Column(
                pn.Row(
                    pn.panel(ICONS["logo"]),
                    pn.Row(pn.Column(self.select.panel, self.control_panel, margin=0), self.description.panel, margin=0),
                ),
                self.plot.panel,
            )

        def setup(self):
            self._setup_watchers()

        @SourceGUI.visible.setter
        def visible(self, visible):
            self._visible = visible

            if visible:
                self.setup()
                self.select.visible = True
                self.description.visible = True
                if len(self.control_panel.objects) == 0:
                    self.control_panel.extend(self.controls)
            elif not visible:
                self.unwatch()
                # do children
                self.select.visible = False
                self.control_panel.clear()
                self.description.visible = False
                self.plot.visible = False
            if self.visible_callback:
                self.visible_callback(visible)

        @property
        def item(self):
            """Item that is selected"""
            if len(self.sources) == 0:
                return None
            return self.sources[0]

except ImportError:

    class GUI(object):
        def __init__(self, *args, **kwargs):
            pass

        def __repr__(self):
            raise RuntimeError("Please install panel to use the GUI (`conda " "install -c conda-forge panel>0.8.0`)")

    EntryGUI = GUI
    CatalogGUI = GUI

except Exception:

    class GUI(object):
        def __init__(self, *args, **kwargs):
            pass

        def __repr__(self):
            raise RuntimeError(
                "Initialization of GUI failed, even though "
                "panel is installed. Please update it "
                "to a more recent version (`conda install -c "
                "conda-forge panel==0.5.1`)."
            )

    EntryGUI = GUI
    CatalogGUI = GUI



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------

from intake.source.base import DataSource


class FooPlugin(DataSource):
    name = "foo"
    version = "0.1"
    container = "dataframe"
    partition_access = False



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------
import argparse
import logging
import signal
import sys

import tornado.ioloop
import tornado.web

from .server import IntakeServer

logger = logging.getLogger("intake")


def call_exit_on_sigterm(signal, frame):
    sys.exit(0)


def main(argv=None):
    from intake import open_catalog
    from intake.config import conf

    if argv is None:
        argv = sys.argv

    parser = argparse.ArgumentParser(description="Intake Catalog Server")
    parser.add_argument("-p", "--port", type=int, default=conf["port"], help="port number for server to listen on")
    parser.add_argument("--list-entries", action="store_true", help="list catalog entries at startup")
    parser.add_argument("--sys-exit-on-sigterm", action="store_true", help="internal flag used during unit testing to ensure " ".coverage file is written")
    parser.add_argument("catalog_args", metavar="FILE", type=str, nargs="+", help="Name of catalog YAML file")
    parser.add_argument("--flatten", dest="flatten", action="store_true")
    parser.add_argument("--no-flatten", dest="flatten", action="store_false")
    parser.add_argument("--ttl", dest="ttl", type=int, default=60)
    parser.add_argument(
        "-a",
        "--address",
        type=str,
        default=conf.get("address", "localhost"),
        help="address to use as a host, defaults to the address " "in the configuration file, if provided otherwise localhost",
    )
    parser.set_defaults(flatten=True)
    args = parser.parse_args(argv[1:])

    if args.sys_exit_on_sigterm:
        signal.signal(signal.SIGTERM, call_exit_on_sigterm)

    logger.info("Creating catalog from:")
    for arg in args.catalog_args:
        logger.info("  - %s" % arg)

    catargs = args.catalog_args
    ttl = args.ttl

    if len(catargs) == 1:
        catalog = open_catalog(catargs[0], ttl=ttl)
        logger.info("catalog_args: %s" % catargs[0])
    else:
        catalog = open_catalog(catargs, flatten=args.flatten, ttl=ttl)
        logger.info("catalog_args: %s" % catargs)
    if args.list_entries:
        # This is not a good idea if the Catalog is huge.
        logger.info("Entries:" + ",".join(list(catalog)))

    logger.info("Listening on %s:%d" % (args.address, args.port))

    server = IntakeServer(catalog)
    app = server.make_app()
    server.start_periodic_functions(close_idle_after=3600.0)

    app.listen(args.port, address=args.address)
    try:
        tornado.ioloop.IOLoop.current().start()
    except KeyboardInterrupt:
        logger.critical("Exiting")
    except Exception as e:
        logger.critical("Exiting due to %s" % e)


if __name__ == "__main__":
    sys.exit(main(sys.argv))



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2019, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------

import os
import shutil

import pytest

import intake

here = os.path.abspath(os.path.dirname(__file__))


@pytest.fixture
def cat1_url():
    return os.path.join(here, "tests", "catalogs", "catalog1.yaml")


@pytest.fixture
def cat2_url():
    return os.path.join(here, "tests", "catalogs", "catalog2.yaml")


@pytest.fixture
def parent_cat_url():
    return os.path.join(here, "tests", "catalogs", "parent.yaml")


@pytest.fixture
def cat1(cat1_url):
    return intake.open_catalog(cat1_url)


@pytest.fixture
def cat2(cat2_url):
    return intake.open_catalog(cat2_url)


@pytest.fixture
def copycat2(cat2_url, tmp_path):
    parent_dir = os.path.dirname(cat2_url)
    shutil.copytree(parent_dir, tmp_path / "catalogs")
    shutil.copytree(os.path.join(os.path.dirname(parent_dir), "data"), tmp_path / "data")
    return intake.open_catalog(tmp_path / "catalogs" / "catalog2.yaml")


@pytest.fixture
def parent_cat(parent_cat_url):
    return intake.open_catalog(parent_cat_url)


@pytest.fixture
def cat_browser(cat1):
    from .catalog.select import CatSelector

    return CatSelector(cats=[cat1])


@pytest.fixture
def sources1(cat1):
    return list(cat1._entries.values())


@pytest.fixture
def sources2(cat2):
    return list(cat2._entries.values())


@pytest.fixture
def source_browser(sources1):
    from .source.select import SourceSelector

    return SourceSelector(sources=sources1)



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------

from .base import DataSource, Schema


class NPySource(DataSource):
    """Read numpy binary files into an array

    Prototype source showing example of working with arrays

    Each file becomes one or more partitions, but partitioning within a file
    is only along the largest dimension, to ensure contiguous data.
    """

    container = "ndarray"
    name = "numpy"
    version = "0.0.1"
    partition_access = True

    def __init__(self, path, dtype=None, shape=None, chunks=None, storage_options=None, metadata=None):
        """
        The parameters dtype and shape will be determined from the first
        file, if not given.

        Parameters
        ----------
        path: str of list of str
            Location of data file(s), possibly including glob and protocol
            information
        dtype: str dtype spec
            In known, the dtype (e.g., "int64" or "f4").
        shape: tuple of int
            If known, the length of each axis
        chunks: int
            Size of chunks within a file along biggest dimension - must exactly divide
            each file, or None for one partition per file.
        storage_options: dict
            Passed to file-system backend.
        """
        self.path = path
        self.shape = shape
        self.dtype = dtype
        self.storage = storage_options or {}
        self.files = None
        self._chunks = chunks
        self._arrs = None
        self._arr = None
        super(NPySource, self).__init__(metadata=metadata)

    def _get_schema(self):
        from fsspec import open_files

        if self.files is None:
            path = self._get_cache(self.path)[0]

            self.files = open_files(path, "rb", compression=None, **self.storage)
            if self.dtype is None or self.shape is None:
                arr = NumpyAccess(self.files[0])
                self.shape = arr.shape
                self.dtype = arr.dtype
            self.chunks = self._chunks or self.shape[0]
        shape = list(self.shape)
        parts_per_file = (shape[0] // self.chunks) + (shape[0] % self.chunks > 0)
        if len(self.files) > 1:
            shape = [len(self.files)] + shape
            chunks = ((1,) * len(self.files),) + ((self.chunks,) * parts_per_file,) + tuple((s,) for s in self.shape[1:])
        else:
            shape = shape
            chunks = ((self.chunks,) * parts_per_file,) + tuple((s,) for s in self.shape[1:])

        print()
        return Schema(dtype=str(self.dtype), shape=tuple(shape), extra_metadata=self.metadata, npartitions=parts_per_file * len(self.files), chunks=chunks)

    def _get_partition(self, i):
        if isinstance(i, (list, tuple)):
            i = i[0]
        parts_per_file = (self.shape[0] // self.chunks) + (self.shape[0] % self.chunks > 0)
        nfile = i // parts_per_file
        arr = NumpyAccess(self.files[nfile])
        infile = i % parts_per_file
        return arr[self.chunks * infile : self.chunks * (infile + 1)]

    def read_partition(self, i):
        return self._get_partition(i)

    def to_dask(self):
        import dask.array as da

        self._get_schema()
        arrs = [NumpyAccess(f, self.shape, self.dtype) for f in self.files]

        if len(arrs) > 1:
            chunks = (self._chunks or -1,) + (-1,) * (len(self.shape) - 1)
        else:
            chunks = (self._chunks or -1,) + (-1,) * (len(self.shape) - 2)
        self._arrs = [da.from_array(arr, chunks) for arr in arrs]

        self._arr = da.stack(self._arrs)
        return self._arr

    def read(self):
        import numpy as np

        self._get_schema()
        with self.files as ff:
            arrs = [np.load(f) for f in ff]
        return np.stack(arrs)

    def _close(self):
        self._arrs = None
        self._arr = None


class NumpyAccess(object):
    def __init__(self, f, shape=None, dtype=None, order="C", offset=None):
        self.f = f
        self.shape = shape
        self.dtype = dtype
        self.order = order
        self.offset = offset
        if self.shape is None or dtype is None or offset is None:
            self._get_info()
        self.ndim = len(self.shape)

    def __getitem__(self, item):
        import copy

        import numpy as np

        if isinstance(item, tuple):
            item = item[0]
        first = (item.stop or self.shape[0]) - (item.start or 0)
        block = item.start or 0
        count = first
        for i in self.shape[1:]:
            block *= i
            count *= i
        if count == 0:
            return np.array([], dtype=self.dtype).reshape(*(-1,) + self.shape[1:])

        start = self.offset + block * self.dtype.itemsize
        shape = (first,) + self.shape[1:]
        fn = copy.copy(self.f)  # makes local copy to avoid close while reading
        with fn as f:
            f.seek(start)
            data = f.read(count * self.dtype.itemsize)
            return np.frombuffer(data, dtype=self.dtype).reshape(shape)

    def _get_info(self):
        from numpy.lib import format

        with self.f as fp:
            version = format.read_magic(fp)
            format._check_version(version)

            shape, fortran_order, dtype = format._read_array_header(fp, version)
            self.shape = shape
            self.dtype = dtype
            self.order = "F" if fortran_order else "C"
            self.offset = fp.tell()



from tiled.client import from_uri
from tiled.client.node import Node

from intake.catalog import Catalog
from intake.source import DataSource


class TiledCatalog(Catalog):
    """View Tiled server as a catalog

    See the documentation for setting up such a server at
    https://blueskyproject.io/tiled/

    A tiled server may contain sources of dataframe, array or xarray type.
    This driver exposes the full tree as exposed by the server, but you
    can also specify the sub-path of that tree.
    """

    name = "tiled_cat"

    def __init__(self, server, path=None):
        """

        Parameters
        ----------
        server: str or tiled.client.node.Node
            Location of tiles server. Usually of the form "http[s]://address:port/"
            May include a path. If the protocol is "tiled", we assume HTTP
            connection. Alternatively, can be a Node instance, already connected
            to a server.
        path: str (optional)
            If given, restrict the catalog to this part of the server's catalog
            tree. Equivalent to extending the server URL.
        """
        self.path = path
        if isinstance(server, str):
            if server.startswith("tiled"):
                uri = server.replace("tiled", "http", 1)
            else:
                uri = server
            client = from_uri(uri, "dask")
        else:
            client = server
            uri = server.uri
        self.uri = uri
        if path is not None:
            client = client[path]
        super().__init__(entries=client, name="tiled:" + uri.split(":", 1)[1])

    def search(self, query, type="text"):
        """Full text search

        Queries other than full text will be added later
        """
        if type == "text":
            from tiled.queries import FullText

            q = FullText(query)
        else:
            raise NotImplementedError
        return TiledCatalog.from_dict(self._entries.search(q), uri=self.uri, path=self.path)

    def __getitem__(self, item):
        node = self._entries[item]
        if isinstance(node, Node):
            return TiledCatalog(node)
        else:
            return TiledSource(uri=self.uri, path=item, instance=node)


types = {
    "DaskArrayClient": "ndarray",
    "DaskDataArrayClient": "xarray",
    "DaskDatasetClient": "xarray",
    "DaskVariableClient": "xarray",
    "DaskDataFrameClient": "dataframe",
}


class TiledSource(DataSource):
    """A source on a Tiled server

    The container type of this source is determined at runtime.
    The attribute ``.instance`` gives access to the underlying Tiled
    API, but most users will only call ``.to_dask()``.
    """

    name = "tiled"

    def __init__(self, uri="", path="", instance=None, metadata=None):
        """

        Parameters
        ----------
        uri: str (optional)
            Location of the server. If ``instance`` is given, this is
            only used for the repr
        pathL str (optional)
            Path of the data source within the server tree. If ``instance``
            is given, this is only used for the repr
        instance: tiled.client.node.None (optional)
            The tiled object pointing to the data source; normally created
            by a ``TiledCatalog``
        metadata: dict
            Extra metadata for this source; metadata will also be provided
            by the server.
        """
        if instance is None:
            instance = from_uri(uri, "dask")[path].read()
        self.instance = instance
        md = dict(instance.metadata)
        if metadata:
            md.update(metadata)
        super().__init__(metadata=md)
        self.name = path
        self.container = types[type(self.instance).__name__]

    def discover(self):
        x = self.to_dask()
        dt = getattr(x, "dtype", None) or getattr(x, "dtypes", None)
        parts = getattr(x, "npartitions", None) or x.data.npartitions
        return dict(dtype=dt, shape=getattr(self.instance.structure().macro, "shape", x.shape), npartitions=parts, metadata=self.metadata)

    def to_dask(self):
        # cache this?
        return self.instance.read()

    def read(self):
        return self.instance.read().compute()

    def _yaml(self):
        y = super()._yaml()
        v = list(y["sources"].values())[0]
        v["args"].pop("instance")
        return y



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2019, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------
import os
from collections import OrderedDict

import panel as pn

MAX_WIDTH = 5000

here = os.path.abspath(os.path.dirname(__file__))
ICONS = {
    "logo": os.path.join(here, "icons", "logo.png"),
    "error": os.path.join(here, "icons", "baseline-error-24px.svg"),
}


def enable_widget(widget, enable=True):
    """Set disabled on widget"""
    widget.disabled = not enable


def coerce_to_list(items, preprocess=None):
    """Given an instance or list, coerce to list.

    With optional preprocessing.
    """
    if not isinstance(items, list):
        items = [items]
    if preprocess:
        items = list(map(preprocess, items))
    return items


class Base(object):
    """
    Base class for composable panel objects that make up the GUI.

    Parameters
    ----------
    children: list of panel objects
        children that will be used to populate the panel when visible
    panel: panel layout object
        instance of a panel layout (row or column) that contains children
        when visible
    watchers: list of param watchers
        watchers that are set on children - cleaned up when visible
        is set to false.
    visible: bool
        whether or not the instance should be visible. When not visible
        ``panel`` is empty.
    logo : bool, opt
        whether to show the intake logo in a panel to the left of the main
        panel. Default is False
    """

    children = None
    panel = None
    watchers = None
    visible_callback = None
    logo_panel = pn.Column(pn.pane.PNG(ICONS["logo"], align="center"), margin=(25, 0, 0, 0), width=50)
    logo = False

    def __init__(self, visible=True, visible_callback=None, logo=False):
        self.visible = visible
        self.visible_callback = visible_callback
        self.logo = logo

    @property
    def panel(self):
        if not self.logo:
            return self._panel
        return pn.Row(self.logo_panel, self._panel, margin=0)

    @panel.setter
    def panel(self, panel):
        self._panel = panel

    def servable(self, *args, **kwargs):
        return self.panel.servable(*args, **kwargs)

    def show(self, *args, **kwargs):
        return self.panel.show(*args, **kwargs)

    def __repr__(self):
        """Print output"""
        return self.panel.__repr__()

    def _repr_mimebundle_(self, *args, **kwargs):
        """Display in a notebook or a server"""
        try:
            return self.panel._repr_mimebundle_(*args, **kwargs)
        except Exception:
            raise NotImplementedError("Panel does not seem to be set " "up properly")

    def setup(self):
        """Should instantiate widgets, set ``children``, and set watchers"""
        raise NotImplementedError

    @property
    def visible(self):
        """Whether or not the instance should be visible."""
        return self._visible

    @visible.setter
    def visible(self, visible):
        """When visible changed, do setup or unwatch and call visible_callback"""
        self._visible = visible
        pan = getattr(self._panel, "_layout", self._panel)
        if visible and len(pan.objects) == 0:
            self.setup()
            self._panel.extend(self.children)
        elif not visible and len(pan.objects) > 0:
            self.unwatch()
            pan.clear()
        if self.visible_callback:
            self.visible_callback(visible)

    def unwatch(self):
        """Get rid of any lingering watchers and remove from list"""
        if self.watchers is not None:
            unwatched = []
            for watcher in self.watchers:
                watcher.inst.param.unwatch(watcher)
                unwatched.append(watcher)
            self.watchers = [w for w in self.watchers if w not in unwatched]

    def __getstate__(self):
        """Serialize the current state of the object"""
        return {"visible": self.visible}

    def __setstate__(self, state):
        """Set the current state of the object from the serialized version.
        Works inplace. See ``__getstate__`` to get serialized version and
        ``from_state`` to create a new object."""
        self.visible = state.get("visible", True)
        return self

    @classmethod
    def from_state(cls, state):
        """Create a new object from a serialized exising object.

        Example
        -------
        original = cls()
        copy = cls.from_state(original.__getstate__())
        """
        return cls().__setstate__(state)


class BaseSelector(Base):
    """Base class for capturing selector logic.

    Parameters
    ----------
    preprocess: function
        run on every input value when creating options
    widget: panel widget
        selector widget which this class keeps uptodate with class properties
    """

    preprocess = None
    widget = None

    @property
    def labels(self):
        """Labels of items in widget"""
        return self.widget.labels

    @property
    def items(self):
        """Available items to select from"""
        return self.widget.values

    @items.setter
    def items(self, items):
        """When setting items make sure widget options are uptodate"""
        if items is not None:
            self.options = items

    def _create_options(self, items):
        """Helper method to create options from list, or instance.

        Applies preprocess method if available to create a uniform
        output
        """
        return OrderedDict(map(lambda x: (x.name, x), coerce_to_list(items, self.preprocess)))

    @property
    def options(self):
        """Options available on the widget"""
        return self.widget.options

    @options.setter
    def options(self, new):
        """Set options from list, or instance of named item

        Over-writes old options
        """
        options = self._create_options(new)
        if self.widget.value:
            self.widget.set_param(options=options, value=list(options.values())[:1])
        else:
            self.widget.options = options
            self.widget.value = list(options.values())[:1]

    def add(self, items):
        """Add items to options"""
        options = self._create_options(items)
        for k, v in options.items():
            if k in self.labels and v not in self.items:
                options.pop(k)
                count = 0
                while f"{k}_{count}" in self.labels:
                    count += 1
                options[f"{k}_{count}"] = v
        self.widget.options.update(options)
        self.widget.param.trigger("options")
        self.widget.value = list(options.values())[:1]

    def remove(self, items):
        """Remove items from options"""
        items = coerce_to_list(items)
        new_options = {k: v for k, v in self.options.items() if v not in items}
        self.widget.options = new_options
        self.widget.param.trigger("options")

    @property
    def selected(self):
        """Value selected on the widget"""
        return self.widget.value

    @selected.setter
    def selected(self, new):
        """Set selected from list or instance of object or name.

        Over-writes existing selection
        """

        def preprocess(item):
            if isinstance(item, str):
                return self.options[item]
            return item

        items = coerce_to_list(new, preprocess)
        self.widget.value = items


class BaseView(Base):
    def __getstate__(self, include_source=True):
        """Serialize the current state of the object. Set include_source
        to False when using with another panel that will include source."""
        if include_source:
            return {
                "visible": self.visible,
                "label": self.source._name,
                "source": self.source.__getstate__(),
            }
        else:
            return {"visible": self.visible}

    def __setstate__(self, state):
        """Set the current state of the object from the serialized version.
        Works inplace. See ``__getstate__`` to get serialized version and
        ``from_state`` to create a new object."""
        if "source" in state:
            self.source = state["source"]
        self.visible = state.get("visible", True)

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, source):
        """When the source gets updated, update the select widget"""
        if isinstance(source, list):
            # if source is a list, get first item or None
            source = source[0] if len(source) > 0 else None
        self._source = source
